/*
 * AAC Audio Data Transport Stream (ADTS) metadata decoder.
 */
#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

typedef struct Framehdr Framehdr;
struct Framehdr {
	uvlong bytes;		/* byte offset in stream */
	uvlong buf;			/* current input buffer */
	uint prev;			/* fixed portion of previous header found */
	ulong num;			/* number of correct headers found */
	ulong sum;			/* sum of audio bytes in valid headers */
	vlong next;			/* expected offset of next header */
	vlong last;			/* offset of last plausable header */

	int vers;			/* Mpeg version */
	int prof;			/* profile */
	int chan;			/* channels */
	double freq;		/* sampling frequency */
	double rate;		/* compraced bit rate */

};

static char *Vers[] = {
	"mp4", "mp2" 
};

static double Freq[] = {
	96, 88.2, 64, 48, 44.1, 32, 24, 22.05, 16,
	12, 11.025, 8, 7.35, -1, -1, -1
};

static char *Prof[] = {
	"aac-main", "aac-lc", "aac-ssr", "aac-ltp"
};

static char *Chan[] = {
	"", "Mono", "Stereo",
	"3.0", "4.0", "5.0", "5.1", "7.1",
	"", "", "", "", "", "", "", ""
};

static uint
val(uvlong n, int off, int len)
{
	return (n >> off) & ((1 << len) -1);
}

static int
validhdr(Framehdr *h)
{
	uint fixed;
	int frmlen;

	if(h->next != 0 && h->bytes != h->next)	/* locked, searching? */
		return h->next - h->bytes;

	fixed = val(h->buf, 28, 28);
	if(fixed != h->prev && h->bytes == h->next){
		h->prev = fixed;
		if(Debug)
			fprint(2, "fixed ");
		goto fail;
	}

	if(val(h->buf, 44, 12) != 0xfff)	/* bad preamble */
		goto fail;
	if(val(h->buf, 41, 2) != 0)			/* bad layer */
		goto fail;
	if(val(h->buf, 38, 2) == 0)			/* bad profile */
		goto fail;
	if(val(h->buf, 34, 4) == 15) 		/* bad sample rate */
		goto fail;
	if(val(h->buf, 33, 1) != 0)			/* bad private flag */
		goto fail;
	if(val(h->buf, 29, 1) != 0)			/* bad origin */
		goto fail;
	if(val(h->buf, 28, 1) != 0)			/* bad home */
		goto fail;
	if(val(h->buf, 23, 3) >= 8)			/* bad channel spec */
		goto fail;

	frmlen = val(h->buf, 13, 13);
	if(h->next == 0){					/* was unlocked? */
		if(Debug)
			Bprint(Meta, "relock, skipped=%lld ", h->bytes - h->last);
		h->next = h->bytes + frmlen;
		return -1;						/* valid header, possible lock */
	}

	h->sum += frmlen;
	h->num++;
	if(h->num > 1000){					/* prevent overflow */
		h->sum /= 2;
		h->num /= 2;
	}

	h->last = h->bytes;
	h->next = h->bytes + frmlen;
	return 0;							/* am locked, got valid header */

fail:
	if(h->bytes == h->next)				/* where we locked? */
		h->next = 0;
	return -1;							/* lost lock */
}


static void
printhdr(Framehdr *h)
{
	double freq, rate;
	int vers, prof, chan;

	freq = Freq[val(h->buf, 34, 4)];
	vers = val(h->buf, 43, 1);
	prof = val(h->buf, 38, 2);
	chan = val(h->buf, 30, 3);

	if(vers != h->vers)
		Bprint(Meta, "wrap %q\n", Vers[vers]);
	if(prof != h->prof)
		Bprint(Meta, "enc %q\n", Prof[prof]);
	if(chan != h->chan)
		Bprint(Meta, "chan %q\n", Chan[chan]);
	if(freq != h->freq)
		Bprint(Meta, "freq %g\n", freq);
	if(h->num > 64){
		rate = 8 * (h->sum / (h->num * 1000.0)) * (freq / 1.024);
		if(fabs(rate - h->rate) > 5){
			Bprint(Meta, "rate %.f\n", rate);
			h->rate = rate;
		}
	}
	Bflush(Meta);

	h->vers = vers;
	h->prof = prof;
	h->chan = chan;
	h->freq = freq;
}

int
adtshdr(int c)
{
	int nxthdr;
	static Framehdr fh;

	if(c == -1){
		memset(&fh, 0, sizeof(fh));
		fh.vers = -1;
		fh.prof = -1;
		fh.chan = -1;
		fh.freq = -1;
		fh.rate = -1;
	}

	fh.bytes++;
	fh.buf = (fh.buf << 8) | (c & 0xff);

	nxthdr = validhdr(&fh);
	if(nxthdr == 0)
		printhdr(&fh);

	/*
	 * we are locked and we are now 
	 * 6 bytes before the end of the header
	 * (i.e. at the start of the header),
	 * we can now start passing data to
	 * audio decoder.
	 */
	if(nxthdr == 6)
		return 1;

	return 0;
}

