#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"

/*
 * We take great steps here to parse mal-formed fields as
 * they are (sadly) rather common
 */

typedef struct Xlate Xlate;
struct Xlate {
	char *src;
	char *dst;
	char *ign;			/* server's default string, ignore this */
	int split;			/* check for "song - artist" format */
};

/*
 * Remember, webfs crushes hyphens out of header names when it generates
 * its virtual filesystem so we need to list header names without
 * hyphens here too.
 */
static Xlate Base[] = {
	{ "icystation",			"station" },
	{ "icygenre",			"genre" },

	{ "icestation",			"station" },
	{ "icegenre",			"genre" },

	{ nil, nil },
};

static Xlate Stream[] = {
	{ "streamtitle",	"title", "",	1 },
	{ "streamurl",		nil, 	 "",	1 },
	{ nil, nil },
};

char *
dequote(char *s)
{
	char *p;

	if(*s != '\'')
		return s;
	s++;
	if((p = strrchr(s, '\'')) != nil)
		*p = 0;
	return s;
}

char *
clean(char *s)
{
	char *p;

	while(isspace(*s))
		s++;
	p = strchr(s, 0) -1;
	while(p >= s && isspace(*p))
		*p-- = 0;
	return s;
};

int
split(char *str, char **arr, int num, char *delim)
{
	char *p;
	int n, len;

	if(Debug)
		fprint(2, "split: '%s'\n", str);

	if((p = strstr(str, delim)) == nil)
		return 0;

	n = 0;
	len = strlen(delim);
	while(p && n < num){
		*p = 0;
		arr[n++] = str;
		str = p + len;
		p = strstr(str, delim);
	}
	if(n < num)
		arr[n++] = str;

	if(Debug)
		fprint(2, "split: -> n=%d\n", n);
	return n;
}

/*
 * expand streamtitle="" metadata which often contain
 * multiple fields.
 */
void
expand(char *str)
{
	int n;
	char *a[8];

	if((n = split(str, a, nelem(a), " - ")) == 0)
		n = split(str, a, nelem(a), " + ");
			

	switch(n){
	default:
		Bprint(Meta, "show %q\n", clean(str));
		break;
	case 2:
		Bprint(Meta, "artist %q\n", clean(a[0]));
		Bprint(Meta, "title %q\n", clean(a[1]));
		break;
	case 3:
		Bprint(Meta, "show %q\n", clean(a[0]));
		Bprint(Meta, "artist %q\n", clean(a[1]));
		Bprint(Meta, "title %q\n", clean(a[2]));
		break;
	}
}
		
	
/*
 * Multi field lines have very poorly defined quoting rules.
 * Whitespace is quoted with single quotes, however there is
 * no quote escape character. We do our best by looking for "';"
 * which often works but we may parse incorrectly somtimes.
 */
static int
multi(char *buf, Xlate *x)
{
	int rc;
	char *nam, *e, *val;

	e = "";
	rc = 0;
	for(nam = buf; *nam; nam = e){
		if((val = strchr(nam, '=')) == nil)
			continue;
		*val++ = 0;

		if(*val == '\''){
			val++;
			if((e = strstr(val, "';")) != nil){		/* 'text';text */
				*e = 0;
				e += 2;
			}
			else
			if((e = strrchr(val, '\'')) != nil)		/* 'text' */
				*e = 0;
			else
				e = strchr(val, 0);					/* 'text (badly formatted) */
		}
		else{
			if((e = strchr(val, ';')) != nil)		/* text;text */
				*e++ = 0;
			else
				e = strchr(val, 0);					/* text */
		}

		for(x = x; x->src; x++)
			if(cistrcmp(nam, x->src) == 0){
				if(x->dst && (x->ign == nil || strcmp(val, x->ign) != 0)){
					if(x->split)
						expand(val);
					else
						Bprint(Meta, "%s %q\n", x->dst, clean(val));
					rc++;
				}
				break;
			}
		if(x->src == nil)
			fprint(2, "icy: '%s=%s' unknown field\n", nam, val);
	}
	return rc;
}

int
icyhdr(Sess *s)
{
	Xlate *x;
	int metaint;
	char val[1024];

	if(httphdr(s, 0, "icymetaint", val, sizeof(val)) < 0){
		werrstr("no icy-metaint: header");
		return -1;
	}

	metaint = atoi(val);
	if(Debug){
		if(metaint == 0)
			Bprint(Meta, "title '<no metadata>'\n");
		else
			Bprint(Meta, "title '<empty>'\n");
	}

	for(x = Base; x->src; x++)
		if(httphdr(s, 0, x->src, val, sizeof(val)) == 0)
			if(x->ign == nil || strcmp(val, x->ign) != 0){
				if(x->dst)
					Bprint(Meta, "%s %q\n", x->dst, clean(dequote(val)));
			}

	Bflush(Meta);
	return metaint;
}

int
icyinband(Biobuf *bp)
{
	int sz;
	char buf[0xff*16];

	if((sz = Bgetc(bp)) == -1)
		return -1;
	if(sz == 0)
		return 0;
	sz *= 16;
	if(Bread(bp, buf, sz) < sz)
		return -1;
	buf[sz] = 0;

	if(Debug)
		if(sz != 0)
			fprint(2, "%s: icy: %d '%s'\n", argv0, sz, buf);

	multi(buf, Stream);
	Bflush(Meta);
	return 0;
}
