#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

int Debug;
int Nosound;
Biobuf *Meta;

/*
 * If the user server name is a browser name (mozilla in it) then
 * you will get stats: e.g. http://stardust.wavestreamer.com:3353/
 *
 * add a semicolon to force the stream version of the data
 * or set the userserver name to "MPEG OVERRIDE"
 *
 * Strangely live365 seems to have problems if we use MPEG OVERRIDE
 * so I tried iTunes instead for a giggle.
 *
 * for V2 servers use:  http://stardust.wavestreamer.com:3353/;?sid=2
 * see
 * http://stackoverflow.com/questions/16871553/how-to-get-html5-to-play-a-stream-from-shoutcast
 *
 * We could start to experiment with shoutcast2 by setting Agent="Ultravox/2.1"
 */
char *Agent = "iTunes";

Metastyle *Metastyles[] = {
	meta_bbc,
	meta_live365
};

typedef struct List List;
struct List {
	char *mimetype;
	char *extname;
	Item *(*parse)(int);
} Playlst[] = {
	{ "audio/x-scpls", 				"pls",		pls },
	{ "audio/x-mpegurl",			"m3u",		m3u },
	{ "application/x-mpegurl",		"m3uu8",	m3u },	/* apple reccomended form */
	{ nil, nil, nil }
};

typedef struct Audio Audio;
struct Audio {
	char *mimetype;
	char *extname;
	int (*framehdr)(int c);
	char *decoder;
} Audiofmt[] = {

	{ "audio/aacp", 			"aac",		adtshdr,	"/bin/audio/adtsdec" },
	{ "audio/aac", 				"aac",		adtshdr, 	"/bin/audio/adtsdec" },
	{ "audio/3gpp",				"3gpp",		adtshdr, 	"/bin/audio/adtsdec" },
	{ "audio/3gpp2",			"3gpp2",	adtshdr, 	"/bin/audio/adtsdec" },

	{ "audio/mpeg", 			"mpeg",		mp3hdr,		"/bin/audio/mp3dec" },
	{ "audio/x-mpeg", 			"mpeg",		mp3hdr,		"/bin/audio/mp3dec" },
	{ "audio/mp2",				"mp2",		mp3hdr,		"/bin/audio/mp3dec" },
	{ "audio/mp3",				"mp3",		mp3hdr,		"/bin/audio/mp3dec" },

	/* maybe add wav and ogg support ... */

	{ nil, nil, nil }
};


int
decompress(char *path)
{
	char *arg[10];
	int narg, fd[2];

	narg = 0;
	arg[narg++] = path;
	arg[narg] = nil;

	if(pipe(fd) < 0){
		fprint(2, "pipe failed %r");
		return -1;
	}

	switch(fork()){
	case -1:
		fprint(2, "fork failed %r");
		close(fd[0]);
		close(fd[1]);
		return -1;
	case 0:
		dup(fd[0], 0);
		close(fd[0]);
		close(fd[1]);
		setprio(getpid(), 16);
		exec(path, arg);
		sysfatal("exec %s failed %r", path);
	default:
		close(fd[0]);
		break;
	}
	return fd[1];
}

int
stream(Sess *s, int net, char *path, int (*framehdr)(int c), Metastyle *helper)
{
	int c, dec;
	Biobuf bnet, bdec;
	vlong last, off, metaint;

	Binit(&bnet, net, OREAD);
	metaint = icyhdr(s);
	last = Boffset(&bnet);

	/* initialise state machine */
	framehdr(-1);

	/* lock to incomming audio frame structure */
	c = -1;
	do{
		if((Boffset(&bnet) - last) == metaint){
			if(icyinband(&bnet) < 0)
				break;
			last = Boffset(&bnet);
		}
		if((c = Bgetc(&bnet)) < 0)
			break;
	}while(! framehdr(c));

	if(Nosound)
		goto done;

	/* start apropriate audio decoder */
	if((dec = decompress(path)) == -1){
		Bterm(&bnet);
		Bflush(Meta);
		return -1;
	}

	/* shuffle bytes, read and strip in-band metadata as needed */
	Binit(&bdec, dec, OWRITE);
	while(1){
		Bputc(&bdec, c);
		off = Boffset(&bnet);

		if((off - last) == metaint){
			if(icyinband(&bnet) < 0)
				break;
			last = Boffset(&bnet);
		}

		if(helper && (off % 8192) == 0)
			helper->sync(helper->arg);

		if((c = Bgetc(&bnet)) < 0)
			break;
		framehdr(c);

	}
fprint(2, "disconnect\n");

	Bterm(&bdec);
	close(dec);

done:
	Bterm(&bnet);
	Bflush(Meta);
	return 0;
}

int
start(Sess *s, int fd, char *audiourl, char *metastyle, char *metaurl)
{
	int i, rc;
	Audio *a, *aud;
	char *ext, *p, buf[128];
	Metastyle *helper;

	httphdr(s, 0, "server", s->server, sizeof(s->server));

	ext = "";
	aud = nil;
	*buf = 0;
	if(httphdr(s, 0, "contenttype", buf, sizeof(buf)) == 0){

		/* can be extra semicolon delimited fields */
		if((p = strchr(buf, ';')) != nil)
			*p = 0;
		for(a = Audiofmt; a->mimetype; a++){
			if(cistrcmp(a->mimetype, buf) == 0){
				aud = a;
				break;
			}
		}
	}

	/* cannot find my mime type, try looking at filename extension */
	if(aud == nil){
		if((p = strrchr(audiourl, '.')) != nil && strlen(p) < 8){
			ext = p+1;
			for(a = Audiofmt; a->mimetype; a++)
				if(cistrcmp(a->extname, ext) == 0){
					aud = a;
					break;
				}
		}
	}

	if(aud == nil){
		werrstr("cannot discern audio format");
		if(Debug)
			fprint(2, "%r - type=%q ext=%q ", buf, ext);
		return -1;
	}

	Bprint(Meta, "server %q\n", s->server);
	Bprint(Meta, "audio %q\n", aud->extname);

	helper = nil;
	if(metastyle != nil && metaurl != nil){
		for(i = 0; i < nelem(Metastyles); i++){
			if(Debug)
				fprint(2, "metastyle=%s name=%s\n", metastyle, Metastyles[i]->name);
			if(strcmp(Metastyles[i]->name, metastyle) == 0)
				helper = Metastyles[i];
		}
	}

	if(helper)
		helper->arg = helper->open(s, metaurl);

	rc = stream(s, fd, aud->decoder, aud->framehdr, helper);

	if(helper)
		helper->close(helper->arg);


	return rc;
}

int
playurl(Sess *s, char *metastyle, char *metaurl, char *audiourl)
{
	int fd, rc;

	if((fd = httpdial(s, 0, Agent, audiourl)) < 0){
		werrstr("playlist ok, cannot get audio - %r");
		return -1;
	}

	if(Debug)
		fprint(2, "%s: %s connected OK\n", argv0, audiourl);

	rc = start(s, fd, audiourl, metastyle, metaurl);
	close(fd);
	return rc;
}

int
listurl(Sess *s, char *audio, char *metastyle, char *metaurl)
{
	int fd, rc;
	Item *items, *i, *t;
	List *l, *lst;
	char *ext, *p, type[128];

	if((fd = httpdial(s, 0, Agent, audio)) < 0){
		werrstr("cannot get playlist - %r");
		return -1;
	}

	if(Debug)
		fprint(2, "%s: %s connected OK\n", argv0, audio);

	lst = nil;
	if(httphdr(s, 0, "contenttype", type, sizeof(type)) == 0){
		if(Debug)
			fprint(2, "%s: %s contenttype=%s\n", argv0, audio, type);

		/* some sites add other semicolon delimited fields */
		if((p = strchr(type, ';')) != nil)
			*p = 0;

		for(l = Playlst; l->mimetype; l++)
			if(cistrcmp(l->mimetype, type) == 0){
				lst = l;
				break;
			}
	}

	/* cannot find my mime type, try looking at filename extension */
	if(lst == nil){
		if((ext = strrchr(audio, '.')) != nil){
			ext++;
			for(l = Playlst; l->mimetype; l++)
				if(cistrcmp(l->extname, ext) == 0){
					lst = l;
					break;
				}
		}
	}

	/* perhaps its a direct link to a stream not a playlist */
	if(lst == nil){
		rc = start(s, fd, audio, metastyle, metaurl);
		close(fd);
		return rc;
	}

	items = (*lst->parse)(fd);
	close(fd);
	if(items == nil){
		werrstr("playlist garbled");
		return -1;
	}

	if(Debug)
		fprint(2, "%s: got items\n", argv0);

	rc = -1;
	for(i = items ;i ; i = i->next){
		if(playurl(s, metastyle, metaurl, i->path) == 0)
			rc = 0;
	}

	for(i = items ;i ; i = t){
		t = i->next;
		free(i->path);
		free(i->title);
		free(i);
	}
	return rc;
}

void
usage(void)
{
	fprint(2, "usage: %s [-n] audio-url [metadata-url]\n", argv0);
	fprint(2, "  -n   do not process audio, only display metadata\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int rc, n, i;
	Biobuf meta;
	Sess sess, *s;
	char *audiourl, *metastyle, *metaurl, buf[128], err[ERRMAX];

	s = &sess;
	Meta = &meta;
	audiourl = nil;
	metastyle = nil;
	metaurl = nil;
	ARGBEGIN{
	case 'n':
		Nosound++;
		break;
	case 'd':
		Debug++;
		break;
	}ARGEND;

	quotefmtinstall();
	Binit(&meta, 1, OWRITE);

	switch(argc){
	case 1:
		audiourl = argv[0];
		break;
	case 3:
		audiourl = argv[0];
		metastyle = argv[1];
		metaurl = argv[2];
		break;
	default:
		usage();
		break;
	}

	s->webfs = "/mnt/web";

	for(i = 0; i < Nconn; i++){
		snprint(buf, sizeof buf, "%s/clone", s->webfs);
		if((s->net[i] = open(buf, ORDWR)) < 0)
			sysfatal("%s cannot open - %r\n", buf);

		if((n = read(s->net[i], buf, sizeof buf-1)) < 0)
			sysfatal("%s cannot read - %r\n", buf);
		if(n == 0)
			sysfatal("%s is empty - %r\n", buf);
		buf[n] = '\0';
		s->conid[i] = atoi(buf);
	}

	setprio(getpid(), 16);

	/*
	 * if our connection drops, and we have found any valid
	 * url in the playlist we must retry. But we must first 
	 * fetch the playlist again, as it may have been updated.
	 * If nothing in the playlist was valid, give up.
	 */
	while((rc = listurl(s, audiourl, metastyle, metaurl)) == 0)
		continue;

	if(rc < 0){
		rerrstr(err, sizeof(err));
		Bprint(Meta, "error %q\n", err);
		rerrstr(buf, sizeof(buf));
		exits(buf);
	}

	for(i = 0; i < Nconn; i++)
		close(s->net[i]);
	exits(nil);
}
