/* Poll the BBC's OnAir info */

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <json.h>
#include "dat.h"
#include "fns.h"

typedef struct {
	int upd;
	char str[256];
} Field;

typedef struct {
	int pid;
	QLock;
	Field show;
	Field title;
	Field artist;
} Onair;

/*
 * BBC's OnAir json feed is of this format:
 *
 * {
 *     "state" : "normal",
 *     "sid" : "bbc_radio_one",
 *     "nowId" : "b03vq0x5",
 *     "nowTitle" : "Fearne Cotton",
 *     "nowLink" : "\/\/www.bbc.co.uk\/showrammes\/b03vq0x5",
 *     "nowInfo" : "A Song About Love by Jake Bugg",
 *     "nextLink" : "\/\/www.bbc.co.uk\/showrammes\/b03vq0x7",
 *     "nextStart" : "2014-03-06T12:45:00+0000",
 *     "nextTitle" : "Newsbeat",
 *     "realtime" : "now_playing"
 * }
 *
 */

static int
upd(Field *f, char *s)
{
	if(strncmp(f->str, s, sizeof(f->str)) != 0){
		snprint(f->str, sizeof(f->str), "%s", s);
		f->upd = 1;
	}
	return 0;
}

static int
split(char *str, char **arr, int num, char *delim)
{
	char *p;
	int n, len;

	if(Debug)
		fprint(2, "split: '%s'\n", str);

	if((p = strstr(str, delim)) == nil)
		return 0;

	n = 0;
	len = strlen(delim);
	while(p && n < num){
		*p = 0;
		arr[n++] = str;
		str = p + len;
		p = strstr(str, delim);
	}
	if(n < num)
		arr[n++] = str;

	if(Debug)
		fprint(2, "split: -> n=%d\n", n+1);
	return n;
}

static void
crack(Onair *oa, char *title, char *info)
{
	int n;
	char *arr[4];

	if(title){
		if(Debug)
			fprint(2, "crack: title='%s'\n", title);

		if((n = split(title, arr, nelem(arr), " by ")) == 0)
			n = split(title, arr, nelem(arr), " - ");
		if(n == 0)
			upd(&oa->show, title);
		if(n == 1)
			upd(&oa->show, arr[0]);
		if(n >= 2){
			upd(&oa->show, arr[0]);
			upd(&oa->title, arr[1]);
		}
	}

	if(info){
		if(Debug)
			fprint(2, "crack: info='%s'\n", info);

		if((n = split(info, arr, nelem(arr), " by ")) == 0)
			n = split(info, arr, nelem(arr), " - ");
		if(n == 0){
			upd(&oa->artist, info);
			upd(&oa->title, "");
		}
		if(n == 1){
			upd(&oa->artist, arr[0]);
			upd(&oa->title, "");
		}
		if(n >= 2){
			upd(&oa->title, arr[0]);
			upd(&oa->artist, arr[1]);
	
			/* show is the first part of info - hapepns on R4 plays */
			if(strcmp(oa->artist.str, oa->show.str) == 0){
				upd(&oa->artist, oa->title.str);
				upd(&oa->title, "");
			}
		}
	}
}

static void
poll(Sess *s, char *url, Onair *oa)
{
	int fd;
	Jvalue *j;
	char etag[1024];

	if(Debug)
		fprint(2, "starting poll of bbc site %s\n", url);

	*etag = 0;
	for(; 1; sleep(30*1000)){	/* The BBC's own javascript polls every 30 secs */

		if((fd = httpredial(s, 1, "9tuner", url, etag)) < 0)
			continue;

		if((j = Jparse(fd, "")) == nil){
			fprint(2, "on-air: json parse failed\n");
			close(fd);
			continue;
		}
		if(httphdr(s, 1, "etag", etag, sizeof(etag)) < 0)
			*etag = 0;
		close(fd);

		qlock(oa);
		crack(oa, Jstrval(j, "nowTitle"), Jstrval(j, "nowInfo"));
		qunlock(oa);

		Jfree(j);
	}
}

static void *
bbc_open(Sess *s, char *url)
{
	Onair *oa;

	if(url == nil || *url == 0)
		return nil;

	if((oa = mallocz(sizeof(Onair), 1)) == nil){
		fprint(2, "No memory %r\n");
		return nil;
	}

	oa->pid = rfork(RFPROC|RFFDG|RFMEM);
	switch(oa->pid){
	case -1:			/* fail */
		free(oa);
		return nil;
	case 0:				/* child */
		poll(s, url, oa);
		exits(0);
	default:			/* parent */
		break;
	}
	return oa;
}

static void
bbc_sync(void *v)
{
	Onair *oa = v;

	if(oa == nil)
		return;

	qlock(oa);
	if(oa->show.upd)
		Bprint(Meta, "show %q\n", oa->show.str);
	if(oa->title.upd)
		Bprint(Meta, "title %q\n", oa->title.str);
	if(oa->artist.upd)
		Bprint(Meta, "artist %q\n", oa->artist.str);
	oa->show.upd = 0;
	oa->title.upd = 0;
	oa->artist.upd = 0;
	qunlock(oa);
}


static void
bbc_close(void *v)
{
	Onair *oa = v;
	Waitmsg *w;

	if(oa == nil)
		return;

	postnote(PNPROC, oa->pid, "die");
	while((w = wait()) != nil)
		if(w->pid == oa->pid)
			break;
	free(w);
	free(oa);
	return;
};

Metastyle meta_bbc[] = {
	.name = "bbc",
	.open = bbc_open,
	.sync = bbc_sync,
	.close = bbc_close,
};
