/*
 * fetch and parse the metadata meant for the Palm7 device.
 * somtimes this is the only way to get the artist/song info
 */
#include <u.h>
#include <libc.h>
#include <bio.h>
#include <xml.h>
#include "dat.h"
#include "fns.h"

struct Lastfm {
	int pid;
	QLock;
	int updated;
	char title[1024];
};

/*
 * last FM rss feed for a user is of this format:
 *
 * <?xml version="1.0" encoding="utf-8"?>
 * <?xml-stylesheet type="text/xsl" href="/rss.xsl" media="screen"?>
 * <rss version="2.0" xmlns:creativeCommons="http://backend.userland.com/creativeCommonsRssModule">
 *    <channel>
 *       <language>en</language>
 *       <creativeCommons:license>http://creativecommons.org/licenses/by-nc-sa/2.0/uk/</creativeCommons:license>
 *             <pubDate>Thu, 13 Feb 2014 13:40:24 +0000</pubDate>
 *       <lastBuildDate>Thu, 13 Feb 2014 13:40:24 +0000</lastBuildDate>
 *             <ttl>5</ttl>
 *       <docs>http://www.audioscrobbler.net/data/webservices</docs>
 *			<title>bbcradio1's Recently Played Tracks</title>
 *       <link>http://www.last.fm/user/bbcradio1</link>
 *       <description>Last 10 tracks submitted to Last.fm</description>
 *       <item>
 *          <title>Tinie Tempah – Lover Not a Fighter (feat. Labrinth)</title>
 *          <link>http://www.last.fm/music/Tinie+Tempah/_/Lover+Not+a+Fighter+(feat.+Labrinth)</link>
 *          <pubDate>Thu, 13 Feb 2014 13:40:24 +0000</pubDate>
 *          <guid>http://www.last.fm/user/bbcradio1#1392298824</guid>
 *                  <description>http://www.last.fm/music/Tinie+Tempah</description>
 *       </item>
 *    </channel>
 * </rss>
 *
 */

static void
poll(Sess *s, char *url, Lastfm *lfm)
{
	int fd;
	Xml *xp;
	Elem *ep;
	char last[128];

	if(Debug)
		fprint(2, "starting poll of rss site %s\n", url);

	*last = 0;
	for(; 1; sleep(60*1000)){
		if((fd = httpredial(s, 1, "9radio", url, last)) < 0)
			continue;

		if((xp = Xparse(fd, 1024*10, Fcrushwhite|Fstripnamespace)) == nil){
			fprint(2, "parse of lastfm RSS feed failed\n");
			close(fd);
			continue;
		}
		httphdr(s, 1, "lastmodified", last, sizeof(last));
		close(fd);

		if((ep = Xfind(xp, xp->root, "rss/channel/item/title")) == nil){
			fprint(2, "lastfm RSS feed not in expected format\n");
			Xfree(xp);
			continue;
		}

		qlock(lfm);
		if(strncmp(lfm->title, ep->pcdata, sizeof(lfm->title)) != 0){
			snprint(lfm->title, sizeof(lfm->title), "%s", ep->pcdata);
			lfm->updated = 1;
		}
		qunlock(lfm);
		
		Xfree(xp);
	}
}

Lastfm *
start_lastfm(Sess *s, char *url)
{
	Lastfm *lfm;

	if(url == nil || *url == 0)
		return nil;

	if((lfm = mallocz(sizeof(Lastfm), 1)) == nil){
		fprint(2, "No memory %r\n");
		return nil;
	}

	lfm->pid = rfork(RFPROC|RFFDG|RFMEM);
	switch(lfm->pid){
	case -1:			/* fail */
		free(lfm);
		return nil;
	case 0:				/* child */
		poll(s, url, lfm);
		exits(0);
	default:			/* parent */
		break;
	}
	return lfm;
}

void
sync_lastfm(Lastfm *lfm)
{
	if(lfm == nil)
		return;

	qlock(lfm);
	if(lfm->updated){
		Bprint(Meta, "rss.title %q\n", lfm->title);
		lfm->updated = 0;
	}
	qunlock(lfm);
}

void
stop_lastfm(Lastfm *lfm)
{
	Waitmsg *w;

	if(lfm == nil)
		return;

	postnote(PNPROC, lfm->pid, "die");
	while((w = wait()) != nil)
		if(w->pid == lfm->pid)
			break;
	free(w);
	free(lfm);
	return;
};
