/*
 * MP3 Transport Stream header decode.
 *
 * Some of the tables where lifted from here:
 * http://www.hydrogenaudio.org/forums/index.php?showtopic=85125&start=0&p=747716&#entry747716
 */

#include <u.h>
#include <libc.h>
#include <bio.h>
#include "dat.h"
#include "fns.h"

typedef struct Framehdr Framehdr;
struct Framehdr {
	uvlong bytes;		/* current input byte counter */
	uvlong next;		/* where we expect the next header */
	uvlong last;		/* last plausable header found */

	int vers;
	int layer;
	int freq;
	int rate;
	int mode;
	int emph;

	uvlong buf;			/* current input buffer */
};

static double Vers[] = {
	2.5, -1, 2, 1 
};

static int Layer[] = {
	-1, 3, 2, 1 
};

static char *Mode[] = { 
	"Stereo", "Joint", "2 Chan", "Mono"
};

static short Rate[4][4][16] = {
	{	// Version 2.5
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 },
		{ 0,   8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, -1 }, // Layer 3
		{ 0,   8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, -1 }, // Layer 2
		{ 0,  32,  48,  56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, -1 }  // Layer 1
	},{	// Reserved
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 },
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 }, // Invalid
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 }, // Invalid
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 }  // Invalid
	},{	// Version 2
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 },
		{ 0,   8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, -1 }, // Layer 3
		{ 0,   8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, -1 }, // Layer 2
		{ 0,  32,  48,  56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, -1 }  // Layer 1
	},{	// Version 1
		{ 0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, -1 },
		{ 0,  32,  40,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, -1 }, // Layer 3
		{ 0,  32,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, -1 }, // Layer 2
		{ 0,  32,  64,  96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1 }, // Layer 1
	}
};

static int Freq[4][4] = {
	{ 11025, 12000,  8000,	-1 },	// v2.5
	{ -1, 	-1, 	-1,		-1 },	// illegal
	{ 22050, 24000, 16000,	-1 },	// v2
	{ 44100, 48000, 32000,	-1 },	// v1
};

static char *Emph[] = {
	"", "15/50ms", "unknown", "itu-j17"
};

static int Samps[4][4] = {
    { -1,  576, 1152,  384 },		// v2.5
    { -1,   -1,   -1,   -1 },		// reserved
    { -1,  576, 1152,  384 },		// v2
    { -1, 1152, 1152,  384 } 		// v1
};

static int Slot[4] = {
	-1,								// reserved
	1,								// layer 3
	1,								// layer 2
	4 								// layer 1
};

static int
val(uvlong n, int off, int len)
{
	return (n >> off) & ((1 << len) -1);
}

static int
validhdr(Framehdr *h)
{
	int frmlen, vers, layer, rate, freq, pad;

	if(h->next != 0 && h->bytes != h->next)	/* locked, but this is not a header */
		return h->next - h->bytes;

	if(val(h->buf, 21, 11) != 0x7ff)		/* frame sync */
		goto fail;
	if((vers = val(h->buf, 19, 2)) == 1)	/* version */
		goto fail;
	if((layer = val(h->buf, 17, 2)) == 0)	/* layer */
		goto fail;
	if((rate = val(h->buf, 12, 4)) == 0x0f) /* bitrate */
		goto fail;
	if(Rate[vers][layer][rate] == -1 || Rate[vers][layer][rate] == 0)
		goto fail;
	if((freq = val(h->buf, 10, 2)) == 3)	/* sample rate */
		goto fail;
	if(val(h->buf, 0, 2) == 2)				/* emphasis */
		goto fail;
	pad = val(h->buf, 9, 1);				/* padding */

	frmlen = (((Samps[vers][layer] / 8) * (double)Rate[vers][layer][rate]*1000.0) /  (double)Freq[vers][freq]);
	if(pad)
		frmlen += Slot[layer];

	if(h->next == 0){
		if(Debug)
			Bprint(Meta, "(%lld)", h->bytes - h->last);
		h->next = h->bytes + frmlen;
		return -1;							/* valid header, maybe relocked */
	}

	h->last = h->bytes;
	h->next = h->bytes + frmlen;
	return 0;								/* locked OK */

fail:
	if(h->bytes == h->next)
		h->next = 0;
	return -1;								/* lost lock */
}


static void
printhdr(Framehdr *h)
{
	int vers, layer, rate, freq, mode, emph;

	vers = val(h->buf, 19, 2);
	layer = val(h->buf, 17, 2);
	rate = val(h->buf, 12, 4);
	freq = val(h->buf, 10, 2);
	mode = val(h->buf, 6, 2);
	emph = val(h->buf, 0, 2);

	if(vers != h->vers)
		Bprint(Meta, "wrap mp%.f\n", Vers[vers]);
	if(layer != h->layer)
		Bprint(Meta, "enc mp%d'\n", Layer[layer]);
	if(vers != h->vers || layer != h->layer || rate != h->rate)
		Bprint(Meta, "rate %d\n", Rate[vers][layer][rate]);
	if(vers != h->vers || freq != h->freq)
		Bprint(Meta, "freq %g\n", Freq[vers][freq] / 1000.0);
	if(mode != h->mode)
		Bprint(Meta, "chan %q\n", Mode[mode]);
	if(emph != h->emph)
		Bprint(Meta, "emph %q\n", Emph[emph]);
	Bflush(Meta);

	h->vers = vers;
	h->layer = layer;
	h->freq = freq;
	h->rate = rate;
	h->mode = mode;
	h->emph = emph;
}

int
mp3hdr(int c)
{
	int nxthdr;
	static Framehdr fh;

	if(c == -1){
		memset(&fh, 0, sizeof(fh));
		fh.vers = -1;
		fh.layer = -1;
		fh.freq = -1;
		fh.rate = -1;
		fh.mode = -1;
		fh.emph = -1;
	}

	fh.buf = (fh.buf << 8) | (c & 0xff);
	fh.bytes++;

	nxthdr = validhdr(&fh);
	if(nxthdr == 0)
		printhdr(&fh);

	/*
	 * we are locked and we are now 
	 * 3 bytes before the end of the header
	 * (i.e. at the start of the header),
	 * we can now start passing data to
	 * audio decoder.
	 */
	if(nxthdr == 3)
		return 1;

	return 0;
}

